
/*
	$Id: mod_dumpcore.c,v 1.4 2003/02/27 12:43:31 anight Exp $

	Copyright (C) 2003 Andrei Nigmatulin <anight at mail dot ru>
	This  program  is free  software;  you can  redistribute  it and/or
	modify  it under the terms of the GNU  General  Public  License  as
	published by the Free Software  Foundation; either version 2 of the
	License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful, but
	WITHOUT  ANY  WARRANTY;   without  even  the  implied  warranty  of
	MERCHANTABILITY  or FITNESS FOR A  PARTICULAR  PURPOSE. See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General  Public  License
	along  with  this  program;  if not,  write  to the  Free  Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* 
	TODO:
		- apply core dump rules on parent apache too
		- directives for limiting core file size
		- directives for limiting total number of core files
		- freebsd port
		- documentation (!)
*/

#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/prctl.h>
#include <stdlib.h>

#include "httpd.h"
#include "http_log.h"
#include "http_config.h"

module MODULE_VAR_EXPORT dumpcore_module;

struct dc_struct {
	int enabled;
	char *dir;
};

static struct dc_struct dc_child_global = {
	enabled: 0,
	dir: NULL
};

static void dc_sigsegv(int c)
{
	char coredir[1024]; /* XXX */
	struct timeval tv;
	struct timezone tz;
	int t;

	if (!dc_child_global.enabled || !dc_child_global.dir) return;

	gettimeofday(&tv, &tz);

	chdir(dc_child_global.dir);

	strftime(coredir, sizeof(coredir), "%Y%m%d-%H%M%S",
		localtime((const time_t *) &tv.tv_sec));

	t = strlen(coredir);
	snprintf(coredir+t, sizeof(coredir)-t, "-%d", getpid());

	mkdir(coredir, 0300);
	chdir(coredir);
}

static void dc_server_init(server_rec *s, pool *p)
{
	/* get default server config */
	struct dc_struct *sconf = ap_get_module_config(s->module_config, &dumpcore_module);
	
	dc_child_global = *sconf;
}

static void *dc_create_server_config(pool *p, server_rec *s)
{
	struct dc_struct *sconf = ap_palloc(p, sizeof(struct dc_struct));

	sconf->enabled = 0;
	sconf->dir = NULL;

	return (void *) sconf;
}

static const char *cmd_dumpcore(cmd_parms *cmd, void *dummy, int flag)
{
	server_rec *s = cmd->server;
	struct dc_struct *sconf = ap_get_module_config(s->module_config, &dumpcore_module);
	sconf->enabled = flag;
	return NULL;
}

static const char *cmd_dumpcoredir(cmd_parms *cmd, void *dummy, char *dir)
{
	server_rec *s = cmd->server;
	struct dc_struct *sconf = ap_get_module_config(s->module_config, &dumpcore_module);
	struct stat st;

	if (stat(dir, &st) < 0) {
		return "cannot stat specified directory";
	}

	if (!S_ISDIR(st.st_mode) || st.st_uid != s->server_uid ||
		(st.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO)) != (S_IWUSR|S_IXUSR)) {

		return "directory must exists, must be owned by apache (see User directive) and have permissions 300";
	}

	sconf->dir = ap_pstrdup(cmd->pool, dir);
	return NULL;
}

static const command_rec dc_cmds[] = {
	{ "DumpCore", cmd_dumpcore, NULL, RSRC_CONF, FLAG,
	  "On or Off to enable or disable saving core dumps" },
	{ "DumpCoreDir", cmd_dumpcoredir, NULL, RSRC_CONF, TAKE1,
	  "Directory where core files should be saved" },
	{NULL}
};

static void dc_child_init(server_rec *s, pool *p)
{
	const struct rlimit r = {
		rlim_cur: RLIM_INFINITY,
		rlim_max: RLIM_INFINITY
	};

	struct sigaction sa;

	/* enable core files */
	if (setrlimit(RLIMIT_CORE, &r) < 0) {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s, "setrlimit() failed");
	}

	/* we have to set mm->dumpable back, after setuid */
	if (prctl(PR_SET_DUMPABLE, 1, 0, 0, 0) < 0) {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s, "prctl() failed");
	}

	memset((void *) &sa, 0, sizeof(sa));
	
	sa.sa_handler = dc_sigsegv;
	/* catch SIGSEGV signal only once.
		this will restore default signal handler on return */
	sa.sa_flags = SA_ONESHOT;

	if (sigaction(SIGSEGV, &sa, NULL) < 0) {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s, "sigaction() failed");
	}
}

module MODULE_VAR_EXPORT dumpcore_module =
{
	STANDARD_MODULE_STUFF,
	dc_server_init,			/* initializer */
	NULL,					/* dir config creater */
	NULL,					/* dir merger --- default is to override */
	dc_create_server_config,/* server config creater*/
	NULL,					/* merge server config */
	dc_cmds,				/* command table */
	NULL,					/* handlers */
	NULL,					/* filename translation */            /* used for wm aliases */
	NULL,					/* check_user_id */
	NULL,					/* check auth */
	NULL,					/* check access */
	NULL,					/* type_checker */
	NULL,					/* fixups */
	NULL,					/* logger */
	NULL,					/* header parser */
	dc_child_init,			/* child_init */
	NULL,					/* child_exit */
	NULL,					/* post read-request */
};
